import numpy as np
from .utils import *
import matplotlib
matplotlib.use('Agg')


class XCal:
    def __init__(self,
                 drug_concentrations: np.ndarray,
                 inhibition_values: np.ndarray,
                 inhibition_sem: np.ndarray = None,
                 concentration_unit: str = 'Nanomolar',
                 drug_name: str = None,
                 cell_line: str = None,
                 experiment: str = None,
                 set_baseline_for_auc: int = 10,
                 set_integration_limit: float = 1,
                 report_quality_scores: bool = False
                 )-> None:

        # initialize meta data
        self._initialize_meta_data(drug_name, cell_line, experiment, concentration_unit)
        # initialize data
        self._initialize_data(drug_concentrations, inhibition_values, inhibition_sem)
        # initialize parameters
        self._initialize_parameters(set_baseline_for_auc, set_integration_limit, report_quality_scores)
        # Updated later
        self._initialize_computational_properties()

    def _initialize_meta_data(self,
                              drug_name: str,
                              cell_line: str,
                              experiment: str,
                              concentration_unit: str
                              ) -> None:
        """
        Initializes metadata of the XCal object.

        :param drug_name: Name of the drug.
        :param cell_line: Name of the cell line.
        :param experiment: Name of the experiment.
        :param concentration_unit: Unit of the drug concentration.
        """
        self.drug_name = drug_name
        self.cell_line = cell_line
        self.experiment = experiment
        self.concentration_unit = concentration_unit

    def _initialize_data(self,
                         drug_concentrations: np.ndarray,
                         inhibition_values: np.ndarray,
                         inhibition_sem: np.ndarray
                         ) -> None:
        """
        Initializes data attributes of the XCal object.

        :param drug_concentrations: Array of drug concentrations.
        :param inhibition_values: Array of inhibition values.
        :param inhibition_sem: Array of standard error of mean for inhibition values.
        """
        # create a df with 'drug_concentration', 'inhibition', 'log10dose' and 'inhibition_sem
        self.df = create_df(drug_concentrations, inhibition_values, inhibition_sem, self.conversion_unit)
        self.drug_concentrations = self.df['drug_concentration'].values
        self.inhibition = self.df['inhibition'].values
        self.log10con = self.df['log10dose'].values
        self.min_log10dose = np.min(np.nanmin(self.log10con))
        self.max_log10dose = np.max(np.nanmax(self.log10con))
        if inhibition_sem is not None:
            self.inhibition_sem = self.df['inhibition_sem'].values
        else:
            self.inhibition_sem = None
        self.max_drug_concentration = np.nanmax(self.drug_concentrations)
        self.min_drug_concentration = np.nanmin(self.drug_concentrations)
        self.number_of_drug_concentrations = len(self.drug_concentrations)
        self.minimum_experimental_inhibition = np.nanmin(self.inhibition)
        self.maximum_experimental_inhibition = np.nanmax(self.inhibition)

    def _initialize_parameters(self,
                               set_baseline_for_auc: int,
                               set_integration_limit: float,
                               report_quality_scores: bool
                               ) -> None:
        """
        Initializes parameters of the XCal object.

        :param set_baseline_for_auc: Baseline value for AUC.
        :param set_integration_limit: Integration limit for calculations.
        :param report_quality_scores: Flag to report quality scores.
        """
        self.set_baseline_for_auc = set_baseline_for_auc
        self.set_integration_limit = set_integration_limit
        self.report_quality_scores = report_quality_scores

    def _initialize_computational_properties(self) -> None:
        """
        Initializes computational properties of the XCal object to None.
        """
        properties = [
            'lmfit_fitted_model', 'error', 'residual', 'log10dose_fine',
            'fitted_curve_fine', 'fitted_values', 'min_val_lm', 'max_val_lm',
            'min_inhibition', 'max_inhibition', 'dose_interpolated',
            'log10_dose_interpolated', 'dose_interpolated_at', 'ic50', 'auc',
            'xepto50', 'r2', 'ar2', 'syx', 'rmse', 'p', 'evar', 'merr', 'rmae',
            'mape', 'area', 'dss1', 'dss2', 'dss3', 'total_area'
        ]

        for prop in properties:
            setattr(self, prop, None)

    # convert concentration unit to value
    @property
    def conversion_unit(self) -> None:
        return unit_conversion(self.concentration_unit)

    def curve_fits(self):
        """
        Perform curve fitting and update relevant properties of the XCal instance.
        """
        self.initial_fit_results = self._initial_curve_fit()
        self._lm_fit(self.initial_fit_results)
        self._generate_fine_curves()
        self._adjust_min_max_inhibition()
        self._calculate_interpolated_values()
        self._calculate_ic50()
        pass

    def _initial_curve_fit(self):
        """
        Perform initial curve fitting using a 4PL model and return the results.
        """
        # use curve fit function to fit initial values
        min_inhibition, max_inhibition, log10ic50_curve_fit, slope_curve_fit = fit_curve_4pl(log10con=self.log10con,
                                                                                             inhibition=self.inhibition
                                                                                             )
        if min_inhibition == max_inhibition:
            max_inhibition += 0.001
        return min_inhibition, max_inhibition, log10ic50_curve_fit, slope_curve_fit


    def _lm_fit(self, initial_fit_results):
        """
        Improve the fitting using parameters from initial curve fitting.
        """
        self.lmfit_fitted_model, lm_fit_results = fit_lm_4pl(log10con=self.log10con,
                                                             inhibition=self.inhibition,
                                                             curve_fitted_values=initial_fit_results
                                                             )
        self.min_val_lm, self.max_val_lm, self.error, self.residual = lm_fit_results


    def _generate_fine_curves(self):
        """
        Generate a dense set of x values over the range of log10dose for a smoother curve.
        """
        self.log10dose_fine = np.linspace(self.min_log10dose, self.max_log10dose, 10000)
        self.fitted_curve_fine = self.lmfit_fitted_model.eval(log10_drug_con=self.log10dose_fine)
        self.fitted_values = self.lmfit_fitted_model.eval(log10_drug_con=self.log10con)
        # print(self.lmfit_fitted_model.params)

    def _adjust_min_max_inhibition(self):
        """
        Adjust minimum and maximum inhibition values.
        """
        min_inhibition = np.min([self.initial_fit_results[0], self.min_val_lm])
        self.min_inhibition = np.max([0, min_inhibition])
        max_inhibition = np.max([self.initial_fit_results[1], self.max_val_lm])
        self.max_inhibition = np.min([100, max_inhibition])


    def _calculate_interpolated_values(self):
        """
        Calculate interpolated values based on fitted curves.
        """
        # Find the index of the value in fitted_curve_fine closest to 50
        if self.min_inhibition <= 50 <= self.max_inhibition:
            index_at_interpolated = np.argmin(np.abs(self.fitted_curve_fine - 50))
            # Extract the corresponding log10dose value at this index
            self.dose_interpolated = ((10 ** self.log10dose_fine[index_at_interpolated]) /
                                      self.conversion_unit).round(2)
            self.dose_interpolated_at = self.fitted_curve_fine[index_at_interpolated].round(1)
        elif self.min_inhibition > 50:
            check_min_fitted_response = np.min(self.fitted_curve_fine)
            if check_min_fitted_response >= 50:
                self.dose_interpolated = ((10 ** np.min(self.log10dose_fine)) / self.conversion_unit).round(2)
                self.dose_interpolated_at = check_min_fitted_response.round(1)
            else:
                index_at_interpolated = np.argmin(np.abs(self.fitted_curve_fine - 50))
                # Extract the corresponding log10dose value at this index
                self.dose_interpolated = ((10 ** self.log10dose_fine[index_at_interpolated]) /
                                          self.conversion_unit).round(2)
                self.dose_interpolated_at = self.fitted_curve_fine[index_at_interpolated].round(1)
        elif self.max_inhibition < 50:
            check_max_fitted_response = np.max(self.fitted_curve_fine)
            if check_max_fitted_response < 50:
                self.dose_interpolated = ((10 ** np.max(self.log10dose_fine)) / self.conversion_unit).round(2)
                self.dose_interpolated_at = check_max_fitted_response.round(1)
            else:
                index_at_interpolated = np.argmin(np.abs(self.fitted_curve_fine - 50))
                self.dose_interpolated = ((10 ** self.log10dose_fine[index_at_interpolated]) /
                                          self.conversion_unit).round(2)
                self.dose_interpolated_at = self.fitted_curve_fine[index_at_interpolated].round(1)
        else:
            the_middle = (self.min_inhibition + self.max_inhibition) * 0.5
            # print(the_middle)
            index_at_interpolated = np.argmin(np.abs(self.fitted_curve_fine - the_middle))
            # Extract the corresponding log10dose value at this index
            self.dose_interpolated = ((10 ** self.log10dose_fine[index_at_interpolated]) /
                                      self.conversion_unit).round(2)
            self.dose_interpolated_at = self.fitted_curve_fine[index_at_interpolated].round(1)
        self.log10_dose_interpolated = np.log10(self.dose_interpolated * self.conversion_unit)


    def _calculate_ic50(self):
        """
        Calculate and adjust IC50 values.
        """
        # convert back to log10ic50 to IC50 of original unit
        self.ic50 = (10 ** self.lmfit_fitted_model.params['log10ic50'].value) / self.conversion_unit

        if self.lmfit_fitted_model.params['slope'].value < 0:
            self.ic50 = self.max_drug_concentration
        if self.lmfit_fitted_model.params['min_val'].value > 90:
            self.ic50 = self.min_drug_concentration

        if np.max([self.minimum_experimental_inhibition, self.maximum_experimental_inhibition]) < 10:
            self.ic50 = self.max_drug_concentration
        if self.minimum_experimental_inhibition > 50:
            self.ic50 = self.min_drug_concentration


    def scores(self):
        """
        Calculate and update various scores of the XCal instance.
        """
        self.auc, _ = cal_auc_aac(fitted_model=self.lmfit_fitted_model,
                                  set_baseline_for_auc=self.set_baseline_for_auc,
                                  max_inhibition=self.max_inhibition,
                                  log10con=self.log10con
                                  )
        self.xepto50 = xepto_score(lmfit_fitted_model=self.lmfit_fitted_model,
                                   dose_interpolated=self.log10_dose_interpolated,
                                   baseline_for_auc=self.set_baseline_for_auc,
                                   max_value_of_inhibition=self.max_inhibition,
                                   integration_limit=self.set_integration_limit
                                   )
        qs = quality_scores(original_values=self.inhibition,
                            fitted_values=self.fitted_values
                            )
        self.r2, self.ar2, self.syx, self.rmse, self.p, self.evar, self.merr, self.rmae, self.mape = qs

        for dss_type in range(1, 4):
            setattr(self, f'dss{dss_type}', self._calculate_dss(dss_type))


    def _calculate_dss(self, dss_type):
        """
        Calculate the DSS score based on the given type.
        """
        return dss(
            ic50=self.ic50,
            slope=self.lmfit_fitted_model.params['slope'].value,
            max_val=self.max_val_lm,
            min_con_tested=self.min_drug_concentration,
            max_con_tested=self.max_drug_concentration,
            y=self.set_baseline_for_auc,
            dss_type=dss_type,
            con_scale=self.conversion_unit
        )

    def generate_plot(self, plot_min, plot_max):
        """
        Generate the plot based on the instance attributes and the given min and max values.

        :param plot_min: Minimum value for plotting.
        :param plot_max: Maximum value for plotting.
        :return: A figure object.
        """
        return xplot(
            ic50=self.ic50,
            dose_interpolated=self.dose_interpolated,
            plot_min=plot_min,
            plot_max=plot_max,
            log10con=self.log10con,
            inhibition=self.inhibition,
            inhibition_sem=self.inhibition_sem,
            log10dose_fine=self.log10dose_fine,
            fitted_curve_fine=self.fitted_curve_fine,
            conversion_unit=self.conversion_unit,
            dose_interpolated_at=self.dose_interpolated_at,
            experiment=self.experiment,
            cell_line=self.cell_line,
            drug_name=self.drug_name,
            set_baseline_for_auc=self.set_baseline_for_auc,
            set_integration_limit=self.set_integration_limit
        )

    def run(self):
        """
        Execute the curve fitting and scoring methods, generate a plot, and compile the results.

        :return: A tuple containing the results dictionary and the figure object.
        """
        self.curve_fits()
        self.scores()
        plot_min = min(self.minimum_experimental_inhibition, self.min_val_lm)
        plot_max = max(self.maximum_experimental_inhibition, self.max_val_lm)
        fig = self.generate_plot(plot_min, plot_max)
        # Define common fields in drug_results
        drug_results = {
            'Experiment': self.experiment,
            'Cell line': self.cell_line,
            'Drug name': self.drug_name,
            'IC50': self.ic50.round(3),
            'IC interpolated': self.dose_interpolated,
            'IC interpolated at': self.dose_interpolated_at,
            'AUC': self.auc,
            'DSS1': self.dss1,
            'DSS2': self.dss2,
            'DSS3': self.dss3,
            'Xepto50': self.xepto50,
        }

        # Add additional fields based on the condition
        if self.report_quality_scores:
            additional_fields = {
                'Minimum drug concentration': self.min_drug_concentration,
                'Maximum drug concentration': self.max_drug_concentration,
                'Number of concentration': self.number_of_drug_concentrations,
                'Concentration unit': self.concentration_unit,
                'Minimum experimental inhibition': self.minimum_experimental_inhibition.round(2),
                'Minimum fitted inhibition': self.lmfit_fitted_model.params['min_val'].value.round(2),
                'Maximum experimental inhibition': self.maximum_experimental_inhibition.round(2),
                'Maximum fitted inhibition': self.max_val_lm.round(2) if self.max_val_lm != 100 else self.max_val_lm,
                'Slope': self.lmfit_fitted_model.params['slope'].value.round(2),
                'IC50 std error': self.error.round(3) if self.error is not None else None,
                'IC50 std residual': self.residual.round(3) if self.residual is not None else None,
                'R2': self.r2,
                'Adjusted R2': self.ar2,
                'RMSE': self.rmse,
                'Normality Test (p)': self.p,
                'Sy.x': self.syx,
                'Explained variance': self.evar,
                'Maximum residual error': self.merr,
                'Mean absolute error': self.rmae,
                'Mean abs percentage error': self.mape
            }
            drug_results.update(additional_fields)

        return drug_results, fig



import numpy as np
from scipy.integrate import quad
from sklearn.metrics import (r2_score, mean_squared_error, explained_variance_score, max_error, mean_absolute_error,
                             mean_absolute_percentage_error)
from scipy import stats
import scipy.integrate as spi
from .curve_fitting import _4pl


def cal_auc_aac(fitted_model,
                set_baseline_for_auc,
                max_inhibition,
                log10con
                ):
    """
    Args:
        fitted_model: a model fitted for curve fitting, here we used lmfit
        set_baseline_for_auc: baseline response
        max_inhibition: maximum response
        log10con: drug concentrations in molar and log10 transformed

    Returns:
        auc: area under the curve
        aac: area above the curve

        The cal_auc_aac function calculates two metrics – Area Under the Curve (AUC) and Above the Curve (AAC) – from
        a fitted 4 Parameter Logistic (4PL) model. These metrics are often used in bioassay and pharmacological studies
        to quantify the overall behavior of the dose-response relationship.
        The function takes the following parameters as inputs:
            fitted_model: An lmfit Model object, presumably fitted using the fit_lm_4pl function described earlier.
            set_baseline_for_auc: The baseline value for calculating AUC.
            max_inhibition: The maximum inhibition value, one of the parameters of the 4PL model.
            log10con: An array or list containing the logarithm base 10 of the drug concentrations.

        The function proceeds with the following steps:
            Define Function for AUC: It defines a function fitted_function_auc that represents the fitted 4PL model
                adjusted by the provided baseline. This function will be used to calculate the AUC.
            Integrate for AUC: Using the quad function from scipy.integrate, it integrates fitted_function_auc over the
                range of log10con to compute the AUC – the area between the adjusted curve and the baseline.
            Define Function for AAC: Similarly, it defines another function fitted_function_aac to represent the maximum
                inhibition value minus the fitted 4PL model. This function will be used to calculate the AAC.
            Integrate for AAC: It integrates fitted_function_aac over the range of log10con to compute the
                AAC – the area above the curve up to the level of maximum inhibition.
            Round and Return Results: Finally, the function rounds the computed AUC and AAC to two decimal places and
            returns them as a tuple.

        By performing these steps, the cal_auc_aac function provides a quantitative assessment of the dose-response
        relationship depicted by the fitted 4PL model, which can be useful for comparing different drug effects or
        assessing drug efficacy and potency.
    """
    # Calculate AUC: Define the function using the fitted parameters, adjusted for the baseline

    def fitted_function_auc(dose_range, baseline):
        function = fitted_model.eval(log10_drug_con=dose_range) - baseline
        return function

    # Integrate over the range of interest to get the area between the curve and the baseline
    integration_result_auc = quad(fitted_function_auc, np.min(log10con), np.max(log10con),
                                  args=(set_baseline_for_auc,))

    # Calculate AAC: Define the function using the fitted parameters, adjusted for the baseline
    def fitted_function_aac(dose_2):
        function = max_inhibition - fitted_model.eval(log10_drug_con=dose_2)
        return function

    # Integrate over the range of interest to get the area between the curve and the baseline
    integration_result_aac = quad(fitted_function_aac, np.min(log10con), np.max(log10con))

    auc = round(integration_result_auc[0], 2)
    aac = round(integration_result_aac[0], 2)

    return auc, aac


def xepto_score(lmfit_fitted_model,
                dose_interpolated,
                baseline_for_auc,
                max_value_of_inhibition,
                integration_limit
                ):
    """
    Args:
        lmfit_fitted_model: Fitted lmfit model.
        dose_interpolated: Interpolated dose value
        baseline_for_auc: Baseline response
        max_value_of_inhibition: Maximum response
        integration_limit: Integration limit

    Returns:

        The xepto_score function calculates a specific score xepto50, based on a fitted model, presumably a 4 Parameter
        Logistic (4PL) model, from lmfit.
        This function is designed to quantify certain areas under the curve within specified limits and baseline
        adjustments.
        The inputs for the function are as follows:

            lmfit_fitted_model: The lmfit Model object that has been fitted with the data.
            dose_interpolated: The starting point for the integration.
            baseline_for_auc: The baseline value for calculating the area under the curve (AUC).
            max_value_of_inhibition: The upper limit used to calculate the area_of_interest.
            integration_limit: The range over which integration is to be performed.
        The function follows the subsequent steps to calculate the score:
            Define AUC Function: The function fitted_function_auc is defined using the parameters of the fitted model,
                adjusted by the baseline provided. It will be used to calculate the areas under the curve.
            Calculate Areas Under the Curve: The function integrates fitted_function_auc over the range specified
                by dose_interpolated and set_integration_limit. The integration_result_auc is adjusted by a fixed
                baseline of 50.
            Calculate Areas of Interest: The area_of_interest is calculated by multiplying set_integration_limit by the
                difference between max_value_of_inhibition and baseline_for_auc.
            Compute Xepto Score: The xepto50 score is computed by dividing the integration
            result by the area_of_interest and multiplying by 100, converting them into percentages.
            Return Results: The function rounds the computed xepto50.
        In essence, the xepto_score function offers a specialized assessment of the fitted model, producing score
        that quantify the proportion of the area under the curve within specific integration limits, with adjustments
        for baseline values. This score can be useful for comparing drug efficacy, potency, or the overall behavior
        of dose-response relationships in different scenarios.
    """
    # Calculate AUC: Define the function using the fitted parameters, adjusted for the baseline
    def fitted_function_auc(dose_range, baseline):
        function = lmfit_fitted_model.eval(log10_drug_con=dose_range) - baseline
        return function

    # Integrate over the range of log10IC50 to get the area between the curve and the baseline
    integration_result_auc = quad(fitted_function_auc, dose_interpolated,
                                    dose_interpolated + integration_limit, args=(50,))

    area_of_interest = (max_value_of_inhibition - baseline_for_auc) * integration_limit
    xepto50 = (integration_result_auc[0] / area_of_interest) * 100

    return round(xepto50, 2)


def quality_scores(original_values,
                   fitted_values
                   ):
    """
    Args:
        original_values: Original response
        fitted_values: Fitted response

    Returns:
        scores: quality control scores

        This quality_scores function is designed to calculate several statistical measures that quantify the quality of
        fit between a set of original (observed) values and a set of model-fitted (predicted) values.
        The function takes two parameters:
            original_values: A list or array containing the original observed values.
            fitted_values: A list or array containing the values predicted by a model.
        The function calculates the following quality scores and returns them as a tuple:
            R² Score (Coefficient of Determination): r2 measures the proportion of the variance in the dependent
                variable that is predictable from the independent variables. It ranges from 0 to 1, with 1 indicating
                perfect prediction.
            Adjusted R² Score: ar2 adjusts the R² score based on the number of observations and the number of
                predictors in the model. It is especially useful when comparing models with different numbers
                of predictors.
            Standard Error of the Estimate (Sy.x): syx provides a measure of the standard deviation of the errors
                (residuals) in the regression model.
            Root Mean Squared Error (RMSE): rmse is a frequently used measure of the differences between values
                predicted by a model and the values observed. It represents the square root of the second sample
                moment of the differences between predicted values and observed values or the quadratic mean of
                these differences.
            Shapiro-Wilk Normality Test P-value: The Shapiro-Wilk test tests the null hypothesis that the data was
                drawn from a normal distribution. A low p-value (< 0.05) indicates that the residuals have a
                distribution that is significantly different from a normal distribution.
            Explained Variance Score: evar measures the proportion to which a mathematical model accounts for the
                variation (dispersion) of a given data set.
            Maximum Residual Error: merr calculates the maximum difference between observed and predicted values.
            Root Mean Absolute Error (RMAE): rmae is the square root of the average of the absolute differences
                between the observed actual outcomes and the predictions made by the model.
            Mean Absolute Percentage Error (MAPE): mape expresses the forecast error as a percentage, which is useful
                when comparing errors across different scales.

        The function then returns a tuple containing all these calculated scores. Each of these scores provides a
        different perspective on the quality of the fit, helping to assess the performance and reliability of the
        regression model.
    """

    r2 = r2_score(fitted_values, original_values)
    # Adjusted R^2
    n = len(original_values)
    k = 1  # Number of predictors
    ar2 = 1 - (1 - r2) * (n - 1) / (n - k - 1)

    # Standard Error of the Estimate (Sy.x)
    syx = np.sqrt(sum((original_values - fitted_values) ** 2) / (n - 2))

    # RMSE
    rmse = np.sqrt(mean_squared_error(original_values, fitted_values))

    # Normality Test (Shapiro-Wilk)
    _, p = stats.shapiro(original_values - fitted_values)

    # Explained variance regression score function.
    evar = explained_variance_score(original_values, fitted_values)

    # The max_error metric calculates the maximum residual error.
    merr = max_error(original_values, fitted_values)

    # Mean absolute error regression loss.
    rmae = np.sqrt(mean_absolute_error(original_values, fitted_values))

    # Mean absolute percentage error (MAPE) regression loss.
    mape = mean_absolute_percentage_error(original_values, fitted_values)

    scores = r2, ar2, syx, rmse, p, evar, merr, rmae, mape

    return scores


def auc_cal(log10_drug_con_min,
            log10_drug_con_max,
            slope,
            set_baseline_for_auc,
            max_val,
            log10ic50
            ):
    """
    Args:
        log10_drug_con_min: Minimum drug concentration in log10 molar
        log10_drug_con_max: Maximum drug concentration in log10 molar
        slope: Hill slope
        set_baseline_for_auc: Baseline response
        max_val: Maximum response
        log10ic50: IC50 in log10 molar

    Returns:
        AUC: area under the curve

        This function auc_cal calculates the Area Under the Curve (AUC) for a 4 Parameter Logistic (4PL) model within a
        specified range of log10 drug concentrations. The 4PL model is often used to describe sigmoidal dose-response
        curves in pharmacology and bioassays. The AUC is a useful metric to quantify the overall effect of a drug.

        Here are the parameters of the function:
            log10_drug_con_min: The minimum log10 drug concentration for calculating the AUC.
            log10_drug_con_max: The maximum log10 drug concentration for calculating the AUC.
            slope: The slope of the curve at the inflection point.
            set_baseline_for_auc: The baseline (minimum response) value for the AUC calculation.
            max_val: The maximum response value of the curve.
            log10ic50: The log10 of the IC50 value, where IC50 is the concentration of the drug that gives half-maximal
                response.
        Within the function:
            The spi.quad function from Scipy is used to numerically integrate (calculate the area under) the 4PL curve
            between log10_drug_con_min and log10_drug_con_max.
            The _4pl function (which should be defined elsewhere in the code) calculates the value of the 4PL model at
            a given log10 drug concentration using the specified parameters (slope, set_baseline_for_auc, max_val,
            log10ic50).
            The quad function returns a tuple where the first element is the calculated AUC, and the second element is
            an estimate of the absolute error in the AUC. In this case, the function returns only the calculated AUC
            (result[0]).

            In summary, this function calculates and returns the AUC of a 4PL dose-response curve between specified
            log10 drug concentration limits using the provided curve parameters.
    """

    result = spi.quad(_4pl, log10_drug_con_min, log10_drug_con_max,
                      args=(slope, set_baseline_for_auc, max_val, log10ic50))

    return result[0]


def dss(ic50, slope, max_val, min_con_tested, max_con_tested, y=10, dss_type=2, con_scale=1e-9):
    a = float(max_val)
    b = float(slope)
    d = 0  # min response
    ic50 = float(ic50)
    min_con_tested = float(min_con_tested)
    max_con_tested = float(max_con_tested)
    min_con = np.log10(min_con_tested * con_scale)
    max_con = max_con_tested
    x2 = np.log10(max_con * con_scale)

    if (np.isnan(ic50) or np.isnan(b) or np.isnan(a) or
            np.isnan(min_con) or np.isnan(max_con)):
        return None

    if ic50 >= max_con:
        return 0

    if b == 0:
        return 0

    if a > 100:
        a = 100

    if b < 0:
        b = -b

    c = np.log10(ic50 * con_scale)

    # This is a logistic function used in Dotmatics.com
    # y = d+(a-d)/(1+10^(b*(c-x)))
    # inverse function
    # x = c - ((log(a-y)-log(d-y))/(b*log(10)))

    if a > y:
        if y != 0:
            x1 = c - ((np.log(a - y) - np.log(y - d)) / (b * np.log(10)))
            if x1 < min_con:
                x1 = min_con
            elif x1 > x2:
                x1 = x2
        else:
            x1 = min_con
        # the first part is integrate from x1 to x2 to get the area under the curve
        # int_y = ((((a - d) * np.log(1 + 10 ** (b * (c - x2)))) / (b * np.log(10)) + a * x2) -
        #         (((a - d) * np.log(1 + 10 ** (b * (c - x1)))) / (b * np.log(10)) + a * x1)) - y * (x2 - x1)

        area = spi.quad(_4pl, x1, x2, args=(slope, d, a, c))
        int_y = area[0]- y * (x2 - x1)

        total_area = (x2 - min_con) * (100 - y)
        norm_area = 0
        if dss_type == 1:
            norm_area = (int_y / total_area) * 100
        elif dss_type == 2:
            norm_area = (int_y / total_area) * 100 / np.log10(a)
            if norm_area > 50:
                norm_area = 0
        elif dss_type == 3:
            norm_area = ((int_y / total_area) * 100 * (np.log10(100) / np.log10(a)) *
                         ((x2 - x1) / (x2 - min_con)))

        if norm_area < 0 or norm_area > 100:
            return 0
        else:
            return round(norm_area, 2)

    else:
        return 0


import pandas as pd
import numpy as np


def unit_conversion(concentration_unit):
    """
    This function takes a preselected name of concentration unit and convert it to a numerical value
    Args:
        concentration_unit: Preselected concentration units.

    Returns:
        conversion_unit
    """

    # Check concentration_unit
    valid_units = ['Molar', 'Millimolar', 'Micromolar', 'Nanomolar', 'Picomolar']
    if concentration_unit not in valid_units:
        raise ValueError(f"Unknown concentration unit: {concentration_unit}")
    conversion_unit = {'Molar': 1, 'Millimolar': 1e-3, 'Micromolar': 1e-6,
                       'Nanomolar': 1e-9, 'Picomolar': 1e-12}[concentration_unit]
    return conversion_unit


def create_df(drug_concentrations, inhibition_values, inhibition_sem=None, conversion_unit=1e-6):
    """
    To create a df and validate data
    Args:
        drug_concentrations: drug concentrations as array or series
        inhibition_values: inhibition values
        inhibition_sem: Standard error of mean for inhibition values, can be None
        conversion_unit: factor to be used to convert drug concentrations

    Returns:
        df: pandas dataframe with three or four columns 'drug_concentration', 'inhibition', 'log10dose'
         and 'inhibition_sem', and sorted by 'drug_concentration'.
    """

    # Check the length of the series and create a dataframe
    if len(drug_concentrations) != len(inhibition_values):
        raise ValueError("Provided concentration data and corresponding response data does not match.")
    df = pd.DataFrame({
        'drug_concentration': drug_concentrations.astype(float),
        'inhibition': inhibition_values.astype(float)
    })
    if inhibition_sem is not None:
        df['inhibition_sem'] = inhibition_sem.astype(float)
    df['log10dose'] = np.log10(df['drug_concentration'] * conversion_unit)
    # sort df by drug_concentration
    df = df.sort_values(by='drug_concentration')
    # Handle duplicated inhibition values
    if df['inhibition'].duplicated().any():
        increment_values = np.arange(0, len(df) * 0.01, 0.01)
        df['inhibition'] += increment_values
    return df


# Function to identify and remove outliers in each row


def remove_row_outliers(row, threshold=1):
    '''

    Args:
        row: A set of data to be used for finding outliers
        threshold: To a value

    Returns:
        row: The data after removal of outliers.

    '''
    Q1 = row.quantile(0.25)
    Q3 = row.quantile(0.75)
    IQR = Q3 - Q1

    # Identify outliers
    outliers = (row < (Q1 - threshold * IQR)) | (row > (Q3 + threshold * IQR))

    # Replace outliers with NaN
    row[outliers] = np.nan
    return row



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('Agg')
# Functions


def xplot(ic50,
          dose_interpolated,
          plot_min,
          plot_max,
          log10con,
          inhibition,
          inhibition_sem,
          log10dose_fine,
          fitted_curve_fine,
          conversion_unit,
          dose_interpolated_at,
          experiment,
          cell_line,
          drug_name,
          set_baseline_for_auc,
          set_integration_limit,
          ):
    """

    Args:
        ic50: Calculated IC50 value
        dose_interpolated: Dose interpolated, interpolated IC50 value
        plot_min: Min response value for plotting
        plot_max: Max response value for plotting
        log10con: Drug concentration log10 in molar
        inhibition: Response values
        inhibition_sem: SEM for response
        log10dose_fine: Log10 doses in molar for fitting, has 10000 values ranging from min log10con to max log10con
        fitted_curve_fine: Fitted response
        conversion_unit: Conversion unit
        dose_interpolated_at: Dose interpolated at concentration
        experiment: Experiment number
        cell_line: Cell line name
        drug_name: Drug name
        set_baseline_for_auc: Baseline response
        set_integration_limit: Set an integration limit for xepto50

    Returns:
        fig: matplotlib figure.

        This function, xplot, is designed to create a plot visualizing a dose-response curve, along with additional
        elements such as data points, interpolated values, IC50, areas under the curve (AUC), and annotations.
        The purpose of such a plot is to visualize how a drug’s inhibition changes with different concentrations and
        to show how well a fitted curve represents the actual data.

        Breakdown of the function:
            Parameters:
                ic50: Concentration of the drug that gives half-maximal response.
                dose_interpolated: The dose level at which interpolation is performed.
                plot_min, plot_max: Define the minimum and maximum for the y-axis scale.
                log10con: Log10 transformed drug concentrations.
                inhibition: Percentage inhibition values corresponding to log10con.
                inhibition_sem: Standard Error of the Mean for inhibition values.
                log10dose_fine: Fine-grained log10 transformed drug concentrations for plotting smooth curves.
                fitted_curve_fine: Fitted curve values corresponding to log10dose_fine.
                conversion_unit: Conversion factor for units of drug concentration.
                dose_interpolated_at: Percentage level at which interpolation is performed.
                experiment: Identifier for the experiment.
                cell_line: Name/Identifier of the cell line used.
                drug_name: Name of the drug.
                set_baseline_for_auc: Baseline value for AUC calculations.
                set_integration_limit: Integration limit for AUC calculations.
        Function Steps:
            Define Fitted Curve Function: Define a helper function fitted_curve that evaluates the lmfit_fitted_model
                at given dose ranges.
            Generate Data for Plotting: Create doses and curve_values arrays for plotting.
            Initialize Plot: Initialize a matplotlib figure and axis object.
            Set Y-Axis Limits and Plot Data Points: Set the limits for the y-axis and plot the actual data points,
                with error bars if inhibition_sem is provided.
            Plot Fitted Curves: Plot the fitted logistic curve and another fitted curve in different colors.
            Draw Vertical Lines and Shade AUC Areas: Draw vertical dashed lines at interpolated dose and log10IC50.
                Shade the area representing xAUC (extended AUC) in red and the regular AUC in light gray.
            Annotate Points: Annotate the interpolated dose and log10IC50 on the plot.
            Set Labels, Title, Legend, and Layout: Set the x-axis and y-axis labels, plot title, add a legend,
                adjust the layout, and close the plot.
            Return Figure: The function returns the matplotlib figure object containing the constructed plot.

        In summary, this function creates a comprehensive visual representation of a dose-response relationship,
        highlighting key parameters, fitted curves, and areas under the curves, which is useful for analyzing the
        efficacy and characteristics of a drug.

    """
    # Plotting
    min_scale = min(0, plot_min-2)
    max_scale = max(100, plot_max+2)
    log10dose_at_interpolated = np.log10(dose_interpolated * conversion_unit)
    log10ic50 = np.log10(ic50 * conversion_unit)

    fig, ax = plt.subplots()
    # Set y-axis limits
    ax.set_ylim([min_scale, max_scale])
    if inhibition_sem is not None:
        ax.errorbar(log10con, inhibition, yerr=inhibition_sem, fmt='o', color='red', label='Actual Data')
    else:
        ax.scatter(log10con, inhibition, color='red', label='Actual Data')
    ax.plot(log10dose_fine, fitted_curve_fine, color='lightblue', label='Fitted Logistic Curve')

    # Draw vertical lines at log10dose_at_50 and log10ic50_est
    ax.axvline(log10dose_at_interpolated, color='green', linestyle='--',
               label=f'Interpolated at {dose_interpolated_at}%')
    ax.axvline(log10ic50, color='blue', linestyle='--', label='IC50')
    """
    # Shade the area under the curve
    if np.max(np.max(fitted_curve_fine)) >=51:
        ax.fill_between(log10dose_fine,
                        fitted_curve_fine,
                        50,
                        where=((log10dose_fine >= log10dose_at_interpolated) &
                               (log10dose_fine <= log10dose_at_interpolated + set_integration_limit)),
                        color='lightgreen', label='xAUC')
    mask = fitted_curve_fine >= set_baseline_for_auc
    log10dose_fine_masked = log10dose_fine[mask]
    fitted_curve_fine_masked = fitted_curve_fine[mask]
    ax.fill_between(log10dose_fine_masked,
                    fitted_curve_fine_masked,
                    set_baseline_for_auc,
                    where=((log10dose_fine_masked >= np.min(log10con)) &
                           (log10dose_fine_masked <= np.max(log10con))),
                    color='lightgray', alpha=0.25, label='AUC')
    """
    # Annotate the points
    min_point = min_scale + 1
    mid_point = min_scale + 45
    ax.annotate(f'Interp@ {dose_interpolated_at}% = {dose_interpolated:.3f}',
                (log10dose_at_interpolated, mid_point), color='green', rotation=90)
    ax.annotate(f'IC50 = {ic50:.3f}', (log10ic50, min_point), color='purple', rotation=90)

    ax.set_xlabel('Log10 Concentration (Molar)')
    ax.set_ylabel('Inhibition %')
    ax.set_title(f'Curve for Exp: {experiment} Cell line: {cell_line} Drug: {drug_name}')
    ax.legend()
    plt.tight_layout()
    plt.close()
    return fig



from .curve_fitting import fit_curve_4pl, fit_lm_4pl
from .utils import unit_conversion, create_df, remove_row_outliers
from .plots import xplot
from .scores import cal_auc_aac, xepto_score, quality_scores, auc_cal, dss



import pandas as pd
import numpy as np
from scipy.optimize import curve_fit
from lmfit import Model, Parameters

# Logistic function from the Hill equation


def _4pl(log10_drug_con, slope, min_val, max_val, log10ic50):
    '''
    Args:
        log10_drug_con: Drug concentration in molar scale, log10 transformed.
        slope: Hill slope
        min_val: Minimum response
        max_val: Maximum response
        log10ic50: Log10 transformed IC50 in molar scale

    Returns:
        response: fitted response

        This is an implementation of the 4 Parameter Logistic (4PL) model, which is commonly used in bioassays and
        dose-response curve fitting. The 4PL model can describe sigmoidal curves, which are often observed in
        biological responses to varying doses of a drug or other substances.
        The parameters of the function:
        log10_drug_con: This parameter represents the logarithm base 10 of the drug concentration.
                        It is a variable on the x-axis in dose-response curves.
        slope: This parameter dictates the steepness of the curve in the sigmoidal 4PL model.
        min_val: This is the minimum value or the lower asymptote of the sigmoidal curve.
        max_val: This is the maximum value or the upper asymptote of the sigmoidal curve.
        log10ic50: This parameter is the logarithm base 10 of the IC50 value.
                   IC50 is a commonly used metric in pharmacology, representing the concentration of a substance
                   required to inhibit a biological process by half.
        The function calculates the response y, given the input parameters, using the 4PL equation:
            y = min_val + (max_val - min_val) / (1 + 10 ** (slope * (log10ic50 - log10_drug_con)))
        The result y is then returned by the function.
        This y value represents the response (on the y-axis of a dose-response curve) corresponding to the given drug
        concentration and the specified curve parameters. The 4PL model and the calculated y are useful for analyzing
        and visualizing the effect of a drug across different concentrations.
    '''

    y = min_val + (max_val - min_val) / (1 + 10 ** (slope * (log10ic50 - log10_drug_con)))
    return y


def fit_curve_4pl(log10con, inhibition):
    '''
    Args:
        log10con: Drug concentration in molar scale, log10 transformed.
        inhibition: A single column response values.

    Returns:
        min_inhibition: The minimum response.
        max_inhibition: The maximum rsponse.
        slope_curve_fit: Hill slope
        log10ic50_curve_fit: Log10 transformed IC50 in molar scale

        This function is designed to fit a set of data points to the 4 Parameter Logistic (4PL) model, making use of
        the _4pl function previously described. It takes as input:
        log10con: An array or list containing the logarithm base 10 of the drug concentrations.
        inhibition: An array or list containing the corresponding inhibition responses.
        The function performs the following operations:
            Initialize Parameters: It sets initial guesses for the parameters of the 4PL model, along with the lower
                and upper bounds for each parameter, to guide the curve fitting.
            Curve Fitting: The function employs the curve_fit method from scipy.optimize to estimate the parameters of
                the 4PL model that best fit the input data. The fitted parameters include the slope, minimum and
                maximum inhibitions, and log10ic50.
            Adjust Minimum and Maximum Inhibition Values: After obtaining the fitted parameters, it makes several
                adjustments to the minimum and maximum inhibition values to ensure they are within reasonable bounds
                (0-99 for minimum inhibition and 1-100 for maximum inhibition).
            Compute Running Mean: It computes the running average of the inhibition values and adjusts the maximum
                inhibition value based on this running average.
            Final Adjustments: It performs several checks and adjustments on the calculated parameters, specifically on
                the log10ic50 value, to ensure it falls within the range of the input log10 concentrations and makes
                sense given the inhibition values.
            Return Fitted Parameters: Finally, it returns the adjusted minimum inhibition, maximum inhibition, slope,
                and log10ic50 values that best describe the input data according to the 4PL model.

            This function is useful for bioassay and pharmacological studies where sigmoidal dose-response curves are
            analyzed to understand the effect of different drug concentrations on biological responses, represented by
            inhibition values in this case. By fitting the data to the 4PL model and obtaining key parameters,
            researchers can gain insights into the characteristics of the drug, such as its potency (IC50) and efficacy
            (maximal effect).
        Use curve_fit from scipy.optimize to estimate parameters, re-runs parameters and covariance
        Initial guesses: [slope, min_val, max_val, log10ic50]
    '''
    min_inhibition_ = np.min(np.nanmin(inhibition))
    max_inhibition_ = np.max(np.nanmax(inhibition))

    upper_min = max(min_inhibition_, 5)
    lower_max = min(max_inhibition_, 95)
    initial_guesses = [1, 0, 100, np.median(log10con)]
    lower_bounds = [0, 0, lower_max, np.min(log10con)]
    upper_bounds = [4, upper_min, 100, np.max(log10con)]

    curve_fit_results, _ = curve_fit(f=_4pl,
                                     xdata=log10con,
                                     ydata=inhibition,
                                     p0=initial_guesses,
                                     bounds=(lower_bounds, upper_bounds),
                                     maxfev=100000)
    # Extract estimated parameters by curve_fit
    slope_curve_fit, min_inhibition_curve_fit, max_inhibition_curve_fit, log10ic50_curve_fit = curve_fit_results
    # print('slope', slope_curve_fit, 'min_val', min_inhibition_curve_fit,
    #       'max_val', max_inhibition_curve_fit, 'logIC50', log10ic50_curve_fit)
    # put a control on min_val_ and min_val_est to set between 0 and 99
    min_inhibition_ori = max(0, min_inhibition_)
    min_inhibition_ori = min(99, min_inhibition_ori)

    min_inhibition_curve_fit = np.nanmax([0, min_inhibition_curve_fit])
    min_inhibition_curve_fit = np.nanmin([99, min_inhibition_curve_fit])
    min_inhibition = min(min_inhibition_ori, min_inhibition_curve_fit)
    # print('min_inhibition:', min_inhibition)

    # put a control on max_val and max_val_est to set between 0 and 100
    max_inhibition_ori = min(100, max_inhibition_)
    max_inhibition_ori = max(1, max_inhibition_ori)

    max_inhibition_curve_fit = np.nanmin([100, max_inhibition_curve_fit])
    max_inhibition_curve_fit = np.nanmax([1, max_inhibition_curve_fit])

    max_inhibition_pre = np.nanmax([max_inhibition_ori, max_inhibition_curve_fit])
    # print('max_inhibition_pre:', max_inhibition_pre)

    # Compute running mean
    running_average = pd.Series(inhibition).rolling(window=3).mean().to_numpy()
    # set the upper value of maximum value
    max_inhibition_run = max_inhibition_pre
    if np.nanmax(running_average[:-1]) > running_average[-1]:
        valid_mask = ~np.isnan(running_average)
        boolean_array = (running_average > running_average[-1]) & valid_mask
        max_inhibition_run = np.nanmax(inhibition[boolean_array])
    if np.nanmax(inhibition) > max_inhibition_run:
        valid_mask = ~np.isnan(inhibition)
        boolean_array = (inhibition > max_inhibition_run) & valid_mask
        max_inhibition_run = np.nanmean(inhibition[boolean_array]) + 1

    max_inhibition_post = np.nanmax([1, max_inhibition_run])
    max_inhibition_post = np.nanmin([100, max_inhibition_post])

    max_inhibition = np.nanmax([max_inhibition_pre, max_inhibition_post])
    # print(min_inhibition, max_inhibition, slope_curve_fit)

    # check if minimum and maximum values qualifies, otherwise set to the experimental values
    if min_inhibition >= max_inhibition:
        log10ic50_curve_fit = np.nanmax(log10con)
    elif log10ic50_curve_fit > np.nanmax(log10con):
        log10ic50_curve_fit = np.nanmax(log10con)
    elif log10ic50_curve_fit < np.nanmin(log10con):
        log10ic50_curve_fit = np.nanmin(log10con)

    if np.nanmax(inhibition) < 0:
        log10ic50_curve_fit = np.nanmax(log10con)
    elif np.nanmin(inhibition) > 100:
        log10ic50_curve_fit = np.nanmin(log10con)

    if np.nanmean(inhibition[:-2]) < 5:
        log10ic50_curve_fit = np.nanmax(log10con)

    return min_inhibition, max_inhibition, slope_curve_fit, log10ic50_curve_fit


def fit_lm_4pl(log10con, inhibition, curve_fitted_values):
    '''

    Args:
        log10con: Drug concentration in molar scale, log10 transformed.
        inhibition: A single column response values.
        curve_fitted_values: Outout from curve fitting.

    Returns:
        lmfit_fitted_model: lmfit model after fitting
        lm_fit_results: results from lmfit model

        This function utilizes the lmfit library to refine the curve fitting of the 4 Parameter Logistic (4PL) model
        on a given set of drug concentration (log10con) and corresponding inhibition values (inhibition). The function
        takes the initial estimates of the parameters (curve_fitted_values) as input, presumably from a previous
        fitting operation (e.g., the fit_curve_4pl function), and returns a more refined model as well as a tuple of
        relevant results.

        A step-by-step description of what this function does:
        Create lmfit Model: Initializes an lmfit Model object based on the previously defined _4pl function.
        Define Initial Parameters: Sets up the initial parameters for the model fitting based on the input
            curve_fitted_values and defined boundaries.
        First Fitting Attempt: Performs the first curve fitting attempt using the initial parameters and evaluates
            the model with the provided log10con and inhibition data.
        Second Fitting Attempt with Adjusted IC50: Adjusts the initial log10ic50 value to the median of log10con and
            performs a second curve fitting attempt.
        Decide on Fitting Result: Compares the standard deviation of residuals from both fitting attempts and chooses
            the model with lower residual standard deviation.
        Adjustment for Low Slope: If the slope value from the chosen model is less than or equal to 0.2, the function
            adjusts the slope bounds and the log10ic50 initial value, and performs another fitting.
        Extract Results and Errors: Extracts relevant results and error information from the fitted model.
        Return Fitted Model and Results: Returns the final fitted model and a tuple containing the minimum value,
            maximum value, error on log10ic50, and standard deviation of residuals.

        By following these steps, the fit_lm_4pl function refines the parameter estimates and achieves a better fit to
        the data using the lmfit library. The returned lmfit model object and the tuple of results can then be used for
        further analysis and interpretation of the biological responses to different drug concentrations.

    '''
    # Create a lmfit model
    min_inhibition, max_inhibition, initial_log10ic50, initial_slope = curve_fitted_values
    # print(min_inhibition, max_inhibition, initial_log10ic50, initial_slope)

    min_max_val = min(90, max_inhibition * 0.9)
    max_min_val = max(10, min_inhibition * 1.1)

    lmfit_model = Model(_4pl)

    # Define initial parameters
    lmfit_params = Parameters()
    lmfit_params.add('slope', value=initial_slope, min=0, max=4)
    # lmfit_params.add('max_val', value=max_val_est, min=lower_max, max=upper_max)
    lmfit_params.add('max_val', value=max_inhibition, min=min_max_val, max=100)
    lmfit_params.add('min_val', value=min_inhibition, min=0, max=max_min_val)
    lmfit_params.add('log10ic50', value=initial_log10ic50, min=np.min(log10con), max=np.max(log10con))

    # Fit the model
    lmfit_fitted_model1 = lmfit_model.fit(data=inhibition, params=lmfit_params, log10_drug_con=log10con)

    # Try with median as starting value for IC50
    lmfit_params['log10ic50'].value = np.median(log10con)
    lmfit_fitted_model2 = lmfit_model.fit(data=inhibition, params=lmfit_params, log10_drug_con=log10con)

    # Decide on which result to use
    ic50std_residual1 = np.std(lmfit_fitted_model1.residual)
    ic50std_residual2 = np.std(lmfit_fitted_model2.residual)
    lmfit_fitted_model = lmfit_fitted_model1 if ic50std_residual1 < ic50std_residual2 else lmfit_fitted_model2
    # print(result_ic50.params)
    # print(result_ic50.params['slope'].value)
    if lmfit_fitted_model.params['slope'].value <= 0.2:
        lmfit_params['log10ic50'].value = initial_log10ic50
        lmfit_params['slope'].min = 0.1
        lmfit_params['slope'].max = 2.5
        lmfit_fitted_model = lmfit_model.fit(data=inhibition, params=lmfit_params, log10_drug_con=log10con)

    error = lmfit_fitted_model.params['log10ic50'].stderr
    residual = np.std(lmfit_fitted_model.residual)
    min_val_lm = lmfit_fitted_model.params['min_val'].value
    max_val_lm = lmfit_fitted_model.params['max_val'].value
    lm_fit_results = min_val_lm, max_val_lm, error, residual
    return lmfit_fitted_model, lm_fit_results



import sys
import pandas as pd
from PyQt5.QtWidgets import (QApplication, QWidget, QVBoxLayout, QPushButton, QFileDialog, QComboBox, QCheckBox,
                             QSpinBox, QLabel, QMainWindow, QAction, QGroupBox, QGridLayout, QSizePolicy, QHBoxLayout,
                             QDoubleSpinBox, QSizePolicy)
from PyQt5.QtGui import QFont
from PyQt5.QtCore import Qt, QThread, pyqtSignal

# Make sure to adjust the import statement according to your project structure
from .run import x50


class XeptoThread(QThread):
    result_signal = pyqtSignal('PyQt_PyObject')
    plots_signal = pyqtSignal('PyQt_PyObject')

    def __init__(self,
                 df,
                 response,
                 response_type,
                 drug_concentration_unit,
                 remove_outlier,
                 set_baseline,
                 set_integration_limit,
                 report_quality_scores
                 ):
        QThread.__init__(self)
        self.df = df
        self.response = response
        self.response_type = response_type
        self.drug_concentration_unit = drug_concentration_unit
        self.remove_outlier = remove_outlier
        self.set_baseline = set_baseline
        self.set_integration_limit = set_integration_limit
        self.report_quality_scores = report_quality_scores

    # def __del__(self):
    #    self.wait()

    def run(self):
        # simulate a long task
        result, plots = x50(df=self.df,
                            response=self.response,
                            response_type=self.response_type,
                            drug_concentration_unit=self.drug_concentration_unit,
                            remove_outlier=self.remove_outlier,
                            set_baseline=self.set_baseline,
                            set_integration_limit=self.set_integration_limit,
                            report_quality_scores=self.report_quality_scores
        )
        self.result_signal.emit(result)  # we use a signal to return the result
        self.plots_signal.emit(plots)


class Window(QMainWindow):
    def __init__(self):
        super(Window, self).__init__()

        # Set window title
        self.setWindowTitle("Xepto50")

        # Create menu bar
        menu = self.menuBar()

        # Create items for the menu bar
        file_menu = menu.addMenu("Menu")

        # Create actions for the file menu
        open_action = QAction("Load", self)
        open_action.triggered.connect(self.load_xlsx)  # connect to the load_csv method
        file_menu.addAction(open_action)

        process_action = QAction("Xepto", self)
        process_action.triggered.connect(self.process)  # connect to the process method
        file_menu.addAction(process_action)

        export_action = QAction("Export", self)
        export_action.triggered.connect(self.export_xlsx)  # connect to the export_csv method
        file_menu.addAction(export_action)

        export_action = QAction("Plots", self)
        export_action.triggered.connect(self.export_plots)  # connect to the export_csv method
        file_menu.addAction(export_action)

        exit_action = QAction("Exit", self)
        exit_action.triggered.connect(self.close)  # connect to the close method
        file_menu.addAction(exit_action)

        # Create central widget
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)

        # Set fixed size
        #self.setFixedSize(600, 500)  # width, height

        # Grid layout
        self.layout = QGridLayout(self.central_widget)  # set layout on central widget


        # Set a title
        self.title = QLabel("Xepto50")
        self.title.setAlignment(Qt.AlignCenter)
        self.title.setStyleSheet("color: lightblue; font-size: 45px; font-weight: bold;")
        self.layout.addWidget(self.title, 0, 0, 1, 2)  # added to row 0, column 0

        self.kazilab = QLabel("KaziLab.se @ Lund University")
        self.kazilab.setAlignment(Qt.AlignRight)
        self.kazilab.setStyleSheet("color: #3399CC; font-size: 10px; font-weight: bold;")
        self.layout.addWidget(self.kazilab, 0, 2, 1, 2)  # added to row 0, column 0

        self.subtitle = QLabel("Calculate drug efficacy in batch mode!")
        self.subtitle.setAlignment(Qt.AlignRight)
        self.subtitle.setStyleSheet("color: #3399CC; font-size: 11px; font-weight: bold;")
        self.layout.addWidget(self.subtitle, 1, 2, 1, 2)  # added to row 0, column 0

        group_box0 = QGroupBox("Data to be used")  # the argument is the title of the group box
        font0 = QFont()
        font0.setPointSize(10)
        font0.setBold(True)
        group_box0.setFont(font0)
        group_box0.setStyleSheet("QGroupBox::title {color: #4a86e8;}")
        # Create a layout for the group box
        group_layout0 = QVBoxLayout()
        # Create an inner QGridLayout
        inner_layout0 = QGridLayout()


        # Label for load csv file button
        self.load_csv_label = QLabel("Load an Excel File:")
        self.load_csv_label.setAlignment(Qt.AlignRight)
        self.load_csv_label.setToolTip('Click "Load Excel file" button to load a data file.')
        self.load_csv_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        #self.layout.addWidget(self.load_csv_label, 1, 0, 1, 1)  # added to row 0, column 0

        # load csv file button
        self.load_csv_button = QPushButton("Load data")
        self.load_csv_button.setStyleSheet("QPushButton { font-size: 12px;}")
        self.load_csv_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.load_csv_button.setMinimumSize(120, 20)  # Set minimum size of QSpinBox (width, height)
        self.load_csv_button.setMaximumSize(120, 20)  # Set maximum size of QSpinBox (width, height)
        self.load_csv_button.setToolTip('Click "Load Excel file" button to load a data file.')
        self.load_csv_button.clicked.connect(self.load_xlsx)
        #self.layout.addWidget(self.load_csv_button, 1, 1, 1, 1)  # added to row 0, column 0, spans 1 row and 2 columns

        # label to display selected file path
        self.file_path_label = QLabel("")
        #self.layout.addWidget(self.file_path_label, 1, 2, 1, 2)  # added to row 1, column 0, spans 1 row and 2 columns

        # Shall we remove outliers
        self.remove_outlier_label = QLabel("Remove outliers:")
        self.remove_outlier_label.setAlignment(Qt.AlignRight)
        self.remove_outlier_label.setToolTip('Check this box if you want to remove outliers')
        self.remove_outlier_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        # self.layout.addWidget(self.mf_for_xgb_label, 3, 0, 1, 1)

        self.remove_outlier_box = QCheckBox()
        self.remove_outlier_box.setStyleSheet(
            "QCheckBox::indicator { width: 20px; height: 20px;} QCheckBox { font-size: 20px;}")
        # self.layout.addWidget(self.mf_for_xgb, 3, 1, 1, 1)
        self.remove_outlier_box.setToolTip('Check this box if you want to remove outliers')

        inner_layout0.addWidget(self.load_csv_label, 0, 0, 1, 1)
        inner_layout0.addWidget(self.load_csv_button, 0, 1, 1, 1)
        inner_layout0.addWidget(self.file_path_label, 0, 2, 1, 2)
        inner_layout0.addWidget(self.remove_outlier_label, 1, 0, 1, 1)
        inner_layout0.addWidget(self.remove_outlier_box, 1, 1, 1, 1)

        # Add inner_layout to the group_layout
        group_layout0.addLayout(inner_layout0)
        # Set the group box's layout
        group_box0.setLayout(group_layout0)
        # Add the group box to the main layout
        self.layout.addWidget(group_box0, 2, 0, 3, 4)

        group_box1 = QGroupBox("Set parameters")  # the argument is the title of the group box
        font1 = QFont()
        font1.setPointSize(10)
        font1.setBold(True)
        group_box1.setFont(font1)
        group_box1.setStyleSheet("QGroupBox::title {color: #4a86e8;}")
        # Create a layout for the group box
        group_layout1 = QVBoxLayout()
        # Create an inner QGridLayout
        inner_layout1 = QGridLayout()

        # response dropdown
        self.response_label = QLabel("Response type:")
        self.response_label.setAlignment(Qt.AlignRight)
        # self.response_label.setToolTip('Set a method to fill missing data with a preliminary number.')
        self.response_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        self.response_label.setToolTip('How drug response was measured, viability or inhibition?')
        # self.layout.addWidget(self.response_label, 2, 0, 1, 1)
        self.response_dropdown = QComboBox()
        self.response_dropdown.setStyleSheet("QComboBox { font-size: 12px;}")
        self.response_dropdown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.response_dropdown.setMinimumSize(120, 20)  # Set minimum size of QSpinBox (width, height)
        self.response_dropdown.setMaximumSize(120, 20)  # Set maximum size of QSpinBox (width, height)
        self.response_dropdown.setToolTip('How drug response was measured, viability or inhibition?')
        self.response_dropdown.addItem("Viability")
        self.response_dropdown.addItem("Inhibition")
        # self.layout.addWidget(self.response_dropdown, 2, 1, 1, 1)  # added to row 1, column 1

        # response dropdown
        self.calculation_label = QLabel("Calculated as:")
        self.calculation_label.setAlignment(Qt.AlignRight)
        # self.calculation_label.setToolTip('Set a method to fill missing data with a preliminary number.')
        self.calculation_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        self.calculation_label.setToolTip('How drug response was calculated, as percentage or ratio?')
        # self.layout.addWidget(self.calculation_label, 2, 0, 1, 1)
        self.calculation_dropdown = QComboBox()
        self.calculation_dropdown.setStyleSheet("QComboBox { font-size: 12px;}")
        self.calculation_dropdown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.calculation_dropdown.setMinimumSize(120, 20)  # Set minimum size of QSpinBox (width, height)
        self.calculation_dropdown.setMaximumSize(120, 20)  # Set maximum size of QSpinBox (width, height)
        self.calculation_dropdown.setToolTip('How drug response was calculated, as percentage or ratio?')
        self.calculation_dropdown.addItem("Percentage")
        self.calculation_dropdown.addItem("Ratio")
        # self.layout.addWidget(self.calculation_dropdown, 2, 1, 1, 1)  # added to row 1, column 1

        # response dropdown
        self.unit_label = QLabel("Concentration unit:")
        self.unit_label.setAlignment(Qt.AlignRight)
        # self.unit_label.setToolTip('Set a method to fill missing data with a preliminary number.')
        self.unit_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        self.unit_label.setToolTip('Unit of drug concentration?')
        # self.layout.addWidget(self.pre_imputation_label, 2, 0, 1, 1)
        self.unit_dropdown = QComboBox()
        self.unit_dropdown.setStyleSheet("QComboBox { font-size: 12px;}")
        self.unit_dropdown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.unit_dropdown.setMinimumSize(120, 20)  # Set minimum size of QSpinBox (width, height)
        self.unit_dropdown.setMaximumSize(120, 20)  # Set maximum size of QSpinBox (width, height)
        self.unit_dropdown.setToolTip('Unit of drug concentration?')
        self.unit_dropdown.addItem("Nanomolar")
        self.unit_dropdown.addItem("Micromolar")
        self.unit_dropdown.addItem("Millimolar")
        self.unit_dropdown.addItem("Molar")
        self.unit_dropdown.addItem("Picomolar")
        # self.layout.addWidget(self.unit_dropdown, 2, 1, 1, 1)  # added to row 1, column 1
        # self.empty1_label = QLabel("")
        # self.empty2_label = QLabel("")
        # self.empty3_label = QLabel("")

        # set_integration_limit spinbox
        self.integration_label = QLabel("Integration limit:")
        self.integration_label.setAlignment(Qt.AlignRight)
        self.integration_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        self.integration_label.setToolTip(f'Set an integration limit to calculate AUC from log10IC50,'
                                          f'Value must be a log10 value, for example 0.1, 0.5, 1 etc.')

        self.integration_spinbox = QDoubleSpinBox()
        self.integration_spinbox.setStyleSheet("QDoubleSpinBox { font-size: 12px;}")
        self.integration_spinbox.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.integration_spinbox.setMinimumSize(110, 20)  # Set minimum size of QDoubleSpinBox (width, height)
        self.integration_spinbox.setMaximumSize(110, 20)  # Set maximum size of QDoubleSpinBox (width, height)
        self.integration_spinbox.setToolTip(f'Set an integration limit to calculate AUC from log10IC50,'
                                            f'Value must be a log10 value, for example 0.1, 0.5, 1 etc.')
        self.integration_spinbox.setRange(0.1, 2)
        self.integration_spinbox.setSingleStep(0.1)
        self.integration_spinbox.setDecimals(2)
        self.integration_spinbox.setValue(1)  # Setting default value to 1

        # Add some widgets to the group box's layout
        inner_layout1.addWidget(self.response_label, 0, 0, 1, 1)
        inner_layout1.addWidget(self.response_dropdown, 0, 1, 1, 1)
        # inner_layout.addWidget(self.empty1_label, 0, 2, 1, 2)

        inner_layout1.addWidget(self.calculation_label, 0, 2, 1, 1)
        inner_layout1.addWidget(self.calculation_dropdown, 0, 3, 1, 1)
        # inner_layout.addWidget(self.empty2_label, 1, 2, 1, 2)

        inner_layout1.addWidget(self.unit_label, 1, 0, 1, 1)
        inner_layout1.addWidget(self.unit_dropdown, 1, 1, 1, 1)
        inner_layout1.addWidget(self.integration_label, 1, 2, 1, 1)
        inner_layout1.addWidget(self.integration_spinbox, 1, 3, 1, 1)

        # Add inner_layout to the group_layout
        group_layout1.addLayout(inner_layout1)
        # Set the group box's layout
        group_box1.setLayout(group_layout1)
        # Add the group box to the main layout
        self.layout.addWidget(group_box1, 5, 0, 3, 4)

        group_box2 = QGroupBox("Set optional parameters")  # the argument is the title of the group box
        font2 = QFont()
        font2.setPointSize(10)
        font2.setBold(True)
        group_box2.setFont(font2)
        group_box2.setStyleSheet("QGroupBox::title {color: #4a86e8;}")
        # Create a layout for the group box
        group_layout2 = QVBoxLayout()
        # Create an inner QGridLayout
        inner_layout2 = QGridLayout()

        # set_baseline spinbox
        self.baseline_label = QLabel("Baseline response:")
        self.baseline_label.setAlignment(Qt.AlignRight)
        self.baseline_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        self.baseline_label.setToolTip(f'Set a baseline response (between 0 and 25, default value: 10)')

        self.baseline_spinbox = QSpinBox()
        self.baseline_spinbox.setStyleSheet("QDoubleSpinBox { font-size: 12px;}")
        self.baseline_spinbox.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.baseline_spinbox.setMinimumSize(110, 20)  # Set minimum size of QDoubleSpinBox (width, height)
        self.baseline_spinbox.setMaximumSize(110, 20)  # Set maximum size of QDoubleSpinBox (width, height)
        self.baseline_spinbox.setToolTip(f'Set a baseline response (between 0 and 25, default value: 10)')
        self.baseline_spinbox.setRange(0, 25)
        self.baseline_spinbox.setSingleStep(5)
        self.baseline_spinbox.setValue(0)  # Setting default value to 0


        # Shall we report quality scores
        self.report_quality_scores_label = QLabel("Fit quality scores:")
        self.report_quality_scores_label.setAlignment(Qt.AlignRight)
        self.report_quality_scores_label.setToolTip('Check this box if you want to report fir quality scores')
        self.report_quality_scores_label.setStyleSheet("QLabel {color: #3399CC; font-size: 12px; font-weight: bold;}")
        # self.layout.addWidget(self.mf_for_xgb_label, 3, 0, 1, 1)

        self.report_quality_scores_box = QCheckBox()
        self.report_quality_scores_box.setStyleSheet(
            "QCheckBox::indicator { width: 20px; height: 20px;} QCheckBox { font-size: 20px;}")
        # self.layout.addWidget(self.mf_for_xgb, 3, 1, 1, 1)
        self.report_quality_scores_box.setToolTip('Check this box if you want to report fir quality scores')

        self.status_label = QLabel("Status: Ready")
        self.status_label.setStyleSheet("color: #3399CC; font-size: 14px; font-weight: bold;")
        # self.layout.addWidget(self.status_label, 9, 2, 1, 2)  # You may want to adjust the position

        # Add some widgets to the group box's layout
        inner_layout2.addWidget(self.baseline_label, 0, 0, 1, 1)
        inner_layout2.addWidget(self.baseline_spinbox, 0, 1, 1, 1)
        # inner_layout.addWidget(self.empty1_label, 0, 2, 1, 2)

        inner_layout2.addWidget(self.report_quality_scores_label, 0, 2, 1, 1)
        inner_layout2.addWidget(self.report_quality_scores_box, 0, 3, 1, 1)

        inner_layout2.addWidget(self.status_label, 1, 1, 1, 2)

        # Add inner_layout to the group_layout
        group_layout2.addLayout(inner_layout2)
        # Set the group box's layout
        group_box2.setLayout(group_layout2)
        # Add the group box to the main layout
        self.layout.addWidget(group_box2, 8, 0, 3, 4)

        # button to handle the function
        self.process_button = QPushButton("Compute")
        self.process_button.clicked.connect(self.process)
        self.process_button.setToolTip('Press to run process - an input file must be provided to start the process')
        self.process_button.setFixedSize(200, 28)  # width, height
        self.process_button.setStyleSheet("QPushButton {font-size: 15px; font-weight: bold; color: #ffffff; background-color: #3399CC;}")
        container = QWidget()
        button_layout = QHBoxLayout()
        button_layout.addWidget(self.process_button, 0, Qt.AlignCenter)
        # Set QHBoxLayout to the container
        container.setLayout(button_layout)
        self.layout.addWidget(container, 12,0,1,2)


        # self.status_label = QLabel("Status: Ready")
        # self.status_label.setStyleSheet("color: #3399CC; font-size: 15px; font-weight: bold;")
        # self.layout.addWidget(self.status_label, 9, 2, 1, 2)  # You may want to adjust the position

        self.export_csv_button = QPushButton("Export result as XLSX")
        self.export_csv_button.clicked.connect(self.export_xlsx)
        self.export_csv_button.setToolTip('Press to export imputed data')
        self.export_csv_button.setFixedSize(200, 28)  # width, height
        self.export_csv_button.setStyleSheet("QPushButton {font-size: 15px; font-weight: bold; color: #ffffff; background-color: lightblue;}")
        container2 = QWidget()
        button_layout2 = QHBoxLayout()
        button_layout2.addWidget(self.export_csv_button, 0, Qt.AlignCenter)
        container2.setLayout(button_layout2)
        self.layout.addWidget(container2, 13,0,1,2)

        self.export_plots_button = QPushButton("Export plots as PDF")
        self.export_plots_button.clicked.connect(self.export_plots)
        self.export_plots_button.setToolTip('Press to export imputed data')
        self.export_plots_button.setFixedSize(200, 28)  # width, height
        self.export_plots_button.setStyleSheet("QPushButton {font-size: 15px; font-weight: bold; color: #ffffff; background-color: lightblue;}")
        container3 = QWidget()
        button_layout3 = QHBoxLayout()
        button_layout3.addWidget(self.export_plots_button, 0, Qt.AlignCenter)
        container3.setLayout(button_layout3)
        self.layout.addWidget(container3, 13,2,1,2)

        # Close button
        self.close_button = QPushButton("Close")
        self.close_button.clicked.connect(self.close)
        self.close_button.setFixedSize(200, 28)  # width, height
        self.close_button.setStyleSheet(
            "font-size: 15px; font-weight: bold; color: #ffffff; background-color: gray")
        container1 = QWidget()
        button_layout1 = QHBoxLayout()
        button_layout1.addWidget(self.close_button, 0, Qt.AlignCenter)
        container1.setLayout(button_layout1)
        self.layout.addWidget(container1, 12,2,1,2)  # You may want to adjust the position

    def load_xlsx(self):
        self.file_name, _ = QFileDialog.getOpenFileName(self, "Open XLSX", "", "XLSX Files (*.xlsx)")
        # update label text with selected file path
        self.file_path_label.setText(f"Loaded file: {self.file_name}")
        if self.file_name:
            print(f"Loaded file {self.file_name}")
        else:
            print("No file was selected")

    def process(self):
        if hasattr(self, 'file_name'):
            df = pd.read_excel(self.file_name)
            response = self.response_dropdown.currentText()
            response_type = self.calculation_dropdown.currentText()
            drug_concentration_unit = self.unit_dropdown.currentText()
            remove_outlier = self.remove_outlier_box.isChecked()
            set_baseline = self.baseline_spinbox.value()
            set_integration_limit = self.integration_spinbox.value()
            report_quality_scores = self.report_quality_scores_box.isChecked()

            self.status_label.setText("Running...")
            self.process_button.setEnabled(False)  # disable the process button while the task is running

            # start the task in a new thread
            self.thread = XeptoThread(df,
                                      response,
                                      response_type,
                                      drug_concentration_unit,
                                      remove_outlier,
                                      set_baseline,
                                      set_integration_limit,
                                      report_quality_scores
                                      )

            # connect signals
            self.thread.result_signal.connect(self.process_finished_result)
            self.thread.plots_signal.connect(self.process_finished_plots)
            self.thread.finished.connect(self.thread.deleteLater)
            self.thread.start()
            # self.thread.wait()

    def process_finished_result(self, result):
        self.status_label.setText("Completed - Ready to export results")
        self.process_button.setEnabled(True)  # re-enable the process button
        self.result = result  # save the result

    def process_finished_plots(self, plots):
        # self.status_label.setText("Completed - Ready to export plots")
        # self.process_button.setEnabled(True)  # re-enable the process button
        self.plots = plots  # save the result

    def export_xlsx(self):
        if hasattr(self, 'result'):
            # This will open a file dialog to select where to save the file
            path, _ = QFileDialog.getSaveFileName(self, "Save XLSX", "", "XLSX Files (*.xlsx)")
            if path:
                if not path.endswith('.xlsx'):
                    path += '.xlsx'
                # export the dataframe to a CSV file
                self.result.to_excel(path)
            else:
                print("No file location provided")
        else:
            print("No dataframe to export")

    def export_plots(self):
        if hasattr(self, 'result'):
            # This will open a file dialog to select where to save the file
            path, _ = QFileDialog.getSaveFileName(self, "Save plots", "", "PDF Files (*.pdf)")
            if path:
                if not path.endswith('.pdf'):
                    path += '.pdf'
                pdf_bytes = self.plots
                with open(path, 'wb') as f:
                    f.write(pdf_bytes)
            else:
                print("No file location provided")
        else:
            print("No plots to export")

if __name__ == "__main__":
    app = QApplication(sys.argv)

    window = Window()
    window.show()

    sys.exit(app.exec_())


def xgui():

    # Create a QApplication instance (needed for any GUI in PyQt)
    app = QApplication([])

    window = Window()
    window.show()

    # Start the application's event loop
    app.exec_()



import pandas as pd
import numpy as np
import matplotlib.backends.backend_pdf
from .cal import XCal
from .utils import remove_row_outliers
import io


def x50(df,
        response='Viability',
        response_type='Percentage',
        drug_concentration_unit='Micromolar',
        remove_outlier=True,
        set_baseline=10,
        set_integration_limit=1,
        report_quality_scores=False
        ):
    '''

    Args:
        df: pandas dataframe
        response: Response upon drug, "Viability" or "Inhibition".
        response_type: How response was canculated "Percentage" or "Ratio".
        drug_concentration_unit: Concentration unit used.
        remove_outlier: Whether outlier to be removed.
        set_baseline: Set a baseline response.
        set_min_con: Set the minimum concentration.
        set_integration_limit: Set a intergation limit.

    Returns:
        final_df: Result df
        pdf_bytes: Figure

    '''
    # Get the list of existing columns
    existing_columns = df.columns.tolist()

    # Create a dictionary to rename the first four columns
    rename_dict = {
        existing_columns[0]: 'experiment_id',
        existing_columns[1]: 'cell_line',
        existing_columns[2]: 'drug_name',
        existing_columns[3]: 'drug_concentration'
    }

    # Rename the columns
    df_renamed = df.iloc[:, :4].rename(columns=rename_dict)

    # Get response columns in a DataFrame
    df_response = df.iloc[:, 4:]
    # print(df_response)
    # Check if single column or multiple columns
    num_columns = len(df_response.columns)
    if num_columns == 0:
        raise ValueError('Response data is missing')
    if response_type == 'Ratio' and np.nanmax(np.nanmax(df_response, axis=1)) <=5:
        df_response = df_response * 100
    elif np.nanmax(np.nanmax(df_response, axis=1)) <= 2:
        df_response = df_response * 100
    if response == 'Viability':
        df_response = 100 - df_response

    if num_columns == 1:
        response_mean = df_response
        response_sem = None
    elif num_columns == 2:
        response_mean = df_response.apply(np.nanmean, axis=1)
        response_sem = None
    else:
        if remove_outlier:
            max_iterations = 3
            previous_sem_sum = float('inf')
            df_imputed = pd.DataFrame()
            for i in range(max_iterations):

                # Remove outliers in each row
                df_outlier_remove = df_response.apply(remove_row_outliers, axis=1)

                # Use row mean to fill in missing values
                df_imputed = df_outlier_remove.apply(lambda row: row.fillna(np.nanmean(row)), axis=1)

                # Recalculate SEM after imputation
                response_sem_post = df_imputed.sem(axis=1)

                # Calculate sum of SEM
                current_sem_sum = response_sem_post.sum()

                if current_sem_sum < previous_sem_sum:
                    previous_sem_sum = current_sem_sum
                    df_response = df_imputed  # Set the imputed df as the new df
                else:
                    break  # Exit the loop if SEM is not reducing

            response_mean = df_imputed.apply(np.nanmean, axis=1)
            response_sem = df_imputed.sem(axis=1)
        else:
            response_mean = df_response.apply(np.nanmean, axis=1)
            response_sem = df_response.sem(axis=1)

    # print(response_sem)
    # df_response.to_excel('df_response.xlsx')
    df_renamed['inhibition_mean'] = response_mean
    if response_sem is not None:
        df_renamed['inhibition_sem'] = response_sem
    # print(df_renamed)
    experiment_ids = df_renamed['experiment_id'].unique()
    dfs = []
    pdf_buffer = io.BytesIO()
    pdf = matplotlib.backends.backend_pdf.PdfPages(pdf_buffer)

    for exp_id in experiment_ids:
        df_id = df_renamed[df_renamed['experiment_id'] == exp_id]
        cell_lines = df_id['cell_line'].unique()
        for line in cell_lines:
            df_line = df_id[df_id['cell_line'] == line]
            drugs = df_line['drug_name'].unique()
            for drug in drugs:
                df_drug = df_line[df_line['drug_name'] == drug]
                if response_sem is not None:
                    drug_results, fig = XCal(drug_concentrations=df_drug['drug_concentration'],
                                             inhibition_values=df_drug['inhibition_mean'],
                                             inhibition_sem=df_drug['inhibition_sem'],
                                             concentration_unit=drug_concentration_unit,
                                             drug_name=drug,
                                             cell_line=line,
                                             experiment=exp_id,
                                             set_baseline_for_auc=set_baseline,
                                             set_integration_limit=set_integration_limit,
                                             report_quality_scores=report_quality_scores
                                             ).run()
                    dfs.append(drug_results)
                    pdf.savefig(fig)
                else:
                    drug_results, fig = XCal(drug_concentrations=df_drug['drug_concentration'],
                                             inhibition_values=df_drug['inhibition_mean'],
                                             inhibition_sem=None,
                                             concentration_unit=drug_concentration_unit,
                                             drug_name=drug,
                                             cell_line=line,
                                             experiment=exp_id,
                                             set_baseline_for_auc=set_baseline,
                                             set_integration_limit=set_integration_limit,
                                             report_quality_scores=report_quality_scores
                                             ).run()
                    dfs.append(drug_results)
                    pdf.savefig(fig)

    # Concatenate all dataframes
    final_df = pd.concat([pd.DataFrame([d]) for d in dfs], ignore_index=True)

    pdf.close()
    # Move the buffer position to the beginning of the stream
    pdf_buffer.seek(0)
    # Read the buffer content to bytes
    pdf_bytes = pdf_buffer.read()
    # Close the buffer
    pdf_buffer.close()
    return final_df, pdf_bytes


'''
# Save the PDF bytes to a file
with open('output_plots.pdf', 'wb') as f:
    f.write(pdf_bytes)
'''


"""
Setup Xepto 50
"""
import sys
try:
    from setuptools import setup, find_packages
except ImportError:
    raise ImportError("Please install `setuptools`")

if not (sys.version_info[0] == 3 and sys.version_info[1] >= 9):
    raise RuntimeError(
                'alphaML requires Python 3.9 or higher.\n'
                'You are using Python {0}.{1}'.format(
                    sys.version_info[0], sys.version_info[1])
                )

# main setup command
setup(
    name='xepto50',
    version='0.0.2', # Major.Minor.Patch
    author='Julhash Kazi',
    author_email='xepto50@kazilab.se',
    url='https://www.kazilab.se',
    description='Calculate drug sensitivity in batch mode!',
    license='Apache-2.0',
    install_requires=[
        'IPython',
        'matplotlib',
        'numpy',
        'scipy',
        'pandas',
        'scikit-learn',
        'lmfit'
    ],
    extras_require={
        'gui': ['PyQt5']
    },
    entry_points={
        'console_scripts': [
            'xepto50=xepto50:main',  # To maps "xepto50" command to the main function
        ],
    },
    platforms='any',
    packages=find_packages()
)



"""
******************
***   Xepto50  ***
******************

Compute efficacy, we adapted xepto from zepto, zeptomolar (zM) is equal to 10^-21 moles/liter .
"""
__author__ = 'Julhash Kazi'
__email__ = 'xepto@kazilab.se'
__description__ = 'Drug efficacy calculation'
__version__ = '0.0.2'
__url__ = 'https://www.kazilab.se'

from .cal import XCal
from .run import x50
from .__main__ import main
